from ..welcsc_payload import _WelcomeScreenChannelPayload
from ..welcomescreen import WelcomeScreenChannel

class WelcomeScreenChannelMapper:
    def __init__(self, dto: _WelcomeScreenChannelPayload):
        self.data = dto
    
    def map(self):
        return WelcomeScreenChannel(self.data.__to_be_conv.data)


"""# DISCMOJI 0.1.0
A discord api wrapper, made for fun.
Pre-alpha, so bugs may be present."""

from guild import *
from ._gateway import DiscordWebsocket
from ._http import HttpManager
from .intents import BotIntents
from .user import User
from .types import *
from .exceptions import *
from .decoration import AvatarDecoration
from .snowflake import Snowflake
from .bot import Bot



from ..ft_payload import _ForumTagsPayload
from ...types import ForumTag

class ForumTagMapper:
    def __init__(self, dto: _ForumTagsPayload):
        self.data = dto
    
    
    def map(self):
        return ForumTag(self.data.__to_be_conv.data)





from ..welcs_payload import _WelcomeScreenPayload
from ..welcomescreen import WelcomeScreen



class WelcomeScreenMapper:
    def __init__(self, dto: _WelcomeScreenPayload):
        self.data = dto

    
    def map(self):
        return WelcomeScreen(self.data.__to_be_conv.data)


from ..s_payload import _StickerPayload
from ..sticker import Sticker



class StickerMapper:
    def __init__(self, dto: _StickerPayload):
        self.data = dto

    
    def map(self):

        return Sticker(self.data.__to_be_conv.data)


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..snowflake import Snowflake
from typing import Optional
from ..types import find_sticker_format_type
from ..user import User
class Sticker:
    """Represents a discord Sticker.
    ## Attributes
    - id - `discmoji.Snowflake`
      - The id of the sticker.
    - pack_id - `discmoji.Snowflake`
      - The id of the pack this sticker came from, if it's a standard sticker.
    - name - `str`
      - The sticker's name.
    - description - `str`
      - The description of the sticker.
    - type - `Literal["STANDARD","GUILD"]`
      - Whether the sticker is part of the guild or is a standard one.
    - format_type - `Literal["PNG","APNG","GIF","LOTTIE"]`
      - The format type of the sticker.
    - available - `Optional[bool]`
      - Whether this guild sticker can be used. May be False due to loss of guild boosts. Can be none.
    - guild_id - `Optional[discmoji.Snowflake]`
      - The id of the guild this sticker came from. Can be none.
    - user - `Optional[discmoji.User]`
      - The user that made this guild sticker. Can be none.
    - sort_value - `Optional[int]`
      - The standard sticker's sort order within a pack.
    """
    def __init__(self,_data: dict):
        self.id = Snowflake(_data["id"])
        self.pack_id: Optional[Snowflake] = Snowflake(_data["pack_id"]) if _data.get("pack_id") else None
        self.name: str = _data["name"]
        self.description: str = _data["description"]
        self.type = "STANDARD" if _data["type"] == 1 else "GUILD"
        self.format_type = find_sticker_format_type(_data["type"])
        self.available: Optional[bool] = _data.get("available")
        self.guild_id: Optional[Snowflake] = Snowflake(_data["guild_id"]) if _data.get("guild_id") else None
        self.user: User = User(_data["user"]) if _data.get("user") else None
        self.sort_value: Optional[int] = _data.get("sort_value")
        
        


from ..e_payload import _EmojiPayload
from ..emoji import Emoji

class EmojiMapper:
    def __init__(self,dto: _EmojiPayload):
        self.data = dto

    
    def map(self):
        return Emoji(self.data.__to_be_conv.data)


from ..vc_payload import _VoiceChannelPayload
from ..channel import VoiceChannel


class VoiceChannelMapper:
    def __init__(self,dto: _VoiceChannelPayload):
        self.data = dto
    
    def map(self):

        return VoiceChannel(self.data.__to_be_conv.data)


from ..fc_payload import _ForumChannelPayload
from ..channel import ForumChannel


class ForumChannelMapper:
    def __init__(self, dto: _ForumChannelPayload):
        self.data = dto
    
    
    def map(self):
        return ForumChannel(self.data.__to_be_conv.data)        


from ..po_payload import _PermissionOverwritePayload
from ..overwrites import PermissionOverwrite

class PermissionOverwriteMapper:
    def __init__(self,dto: _PermissionOverwritePayload):
        self.data = dto
    
    
    def map(self):
        return PermissionOverwrite(self.data.__to_be_conv.data)


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..snowflake import Snowflake
from typing import Optional,TypeAlias
from ..types import _find_verification_level,_find_notif_level,_find_expl_level,_get_system_flags,_get_nitro_rank,Locales
from .roles import Role
from .emoji import Emoji
from .welcomescreen import WelcomeScreen
from .sticker import Sticker
from ..exceptions import DiscmojiRetrievalError,Forbidden,UnknownHTTPError
from ..bot import Bot
from mappers.gp_mapper import GuildPreviewMapper
from .gp_payload import _GuildPreviewPayload
from .channel import Channel
from .c_payload import _ChannelPayload
from mappers.c_mapper import ChannelMapper
class Guild:
    """Represents a Guild/Server on Discord. 
    There is an alias for this called Server.
    ## Attributes
    - id - `discmoji.Snowflake`
       - Contains the id of the guild.
    - name - `str`
       - Contains the name of the guild.
    - icon - `str`
       - Contains the formatted CDN link of the guild's icon.
    - splash - `str`
       - Contains the formatted CDN link of the guild's splash.
    - discovery_splash - `Optional[str]`
       - Contains the formatted CDN link of the guild's discovery splash, if it is eligible and the option is enabled.
    - owner - `Optional[bool]`
       - Only a non-None value if this guild was constructed through the `Get User Guilds` endpoint.
    - permissions - `Optional[str]`
       - Only a non-None value if this guild was constructed through the `Get User Guilds` endpoint.
    - afk_channel_id - `Optional[discmoji.Snowflake]`
       - Contains the channel id in which guild members are moved to when the pass the inactivity threshold.
    - afk_timeout - `int`
       - Contains the length of the timeout if the guild member has passed the inactivity threshold.
    - widget_enabled - `Optional[bool]`
       - Indicates whether a widget (invite) has been enabled on this guild. May be None.
    - widget_channel_id - `Optional[discmoji.Snowflake]`
       - Contains the channel id in which the guild's widget directs to. May be None.
    - verification_level - `dict[str,int]`
       - Contains a dict with one key, which can be one of 5 string literals: `NONE`,`LOW`,`MEDIUM`,`HIGH`,`VERY_HIGH`, and whose value corresponds an int in the range of 0 to 4.
    - default_msg_notifs - `dict[str,int]`
       - Contains a dict with one key, which can be one of 2 string literals: `ALL_MESSAGES`,`ONLY_MENTIONS`, and whose value corresponds to an int in the range of 0 to 1.
    - explicit_filter_lvl - `dict[str,int]`
       - Contains a dict with one key, which can be one of 3 string literals: `DISABLED`,`MEMBERS_WITHOUT_ROLES`,`ALL_MEMBERS` and whose value corresponds an int in the range of 0 to 2.
    - roles - `list[discmoji.Role]`
       - Contains a list of all the guild's roles.
    - emojis - `list[discmoji.Emoji]`
       - Contains a list of all the guild's emojis.
    - features - `list[str]`
       - List of strings listing enabled guild features. 
    - mfa_level - `Literal["NONE","ELEVATED"]`
      - String literal indicating **M**ulti **F**actor **A**uthentication level to enter guild.
    - application_id - `Optional[discmoji.Snowflake]`
      - Contains the application's id if the guild is bot creates, else None.
    - system_channel_id - `Optional[discmoji.Snowflake]`
      - Contains the system channel's id, if exists.
    - system_channel_flags - `list[dict[str,int]]`
      - A list of all enabled system channel flags.
      - key is always capital ( [see ddev documentation](https://discord.com/developers/docs/resources/guild#guild-object-system-channel-flags) ), value is always an integer.
    - rules_channel_id - `Optional[discmoji.Snowflake]`
      - The rule channel's id, if exists.
    - max_presences - `Optional[int]`
      - Almost always None, unless guild has a very sizable amount of members.
        - e.g Midjourney
    - max_members - `Optional[int]`
        - The maximum member threshold for the guild.
    - vanity_url_code - `Optional[str]`
        - The vanity url code for the guild.
    - description - `Optional[str]`
        - The description of the guild.
    - banner - `Optional[str]`
        - Contains a CDN link to the guild's banner. May be None.
    - premium_tier - `Literal["Tier 1","Tier 2","Tier 3"] | None`
        - The server boost rank of the guild.
    - premium_subscription_count - `int`
        - The amount of boosts this guild has.
    - preferred_locale - `str`
        - Returns the language for the locale this guild is in.
    - public_upd_channel_id - `Optional[discmoji.Snowflake]`
        - the public updates channel's id, if exists.
    - max_video_channel_users - `Optional[int]`
        - The maximum amount of users that can be inside a video channel.
    - max_stage_video_channel_users - `Optional[int]`
        - The maximum amount of users that can be inside a stage video channel.
    - approximate_member_count - `Optional[int]`
        - Only a non-None value if retrieved from the `Get Guild` endpoint with the `with_count` param.
    - approximate_presence_count - `Optional[int]`
        - Only a non-None value if retrieved from the `Get Guild` endpoint with the `with_count` param.
    - welcome_screen - `Optional[WelcomeScreen]`
        - The welcome screen for this guild, if exists.
    - stickers - `Optional[list[discmoji.Sticker]]`
        - A list of stickers for this guild.
    - progress_bar_enabled - `bool`
        - Whether the guild's boost progress bar is enabled.
    - safety_alerts_channel_id - `Optional[discmoji.Snowflake]`
        - The id of the safety alerts channel for this guild."""
    def __init__(self,_data: dict[str, str | int | dict | None | bool]):
        self.id = Snowflake(_data["id"])
        self.name: str = _data["name"]
        self.icon: str = f"https://cdn.discordapp.com/icons/{self.id}/{_data["icon"]}.{"gif" if _data["icon"].startswith("a_") else "png"}"
        self.splash: str = f"https://cdn.discordapp.com/splashes/{self.id}/{_data["splash"]}.{"gif" if _data["splash"].startswith("a_") else "png"}"
        self.discovery_splash: Optional[str] = f"https://cdn.discordapp.com/discovery-splashes/{self.id}/{_data.get("discovery_splash")}.{"gif" if _data.get("discovery_splash").startswith("a_") else "png"}" if _data.get("discovery_splash") is not None else None
        self.owner: Optional[bool] = _data.get("owner")
        self.owner_id: Snowflake = Snowflake(_data["owner_id"])
        self.permissions: Optional[str] = _data.get("permissions")
        self.afk_channel_id: Optional[Snowflake] = Snowflake(_data.get("afk_channel_id")) if _data.get("afk_channel_id") is not None else None
        self.afk_timeout: int = _data["afk_timeout"]
        self.widget_enabled: Optional[bool] = _data.get("widget_enabled")
        self.widget_channel_id: Optional[Snowflake] = Snowflake(_data.get("widget_channel_id")) if _data.get("widget_channel_id") is not None else None
        self.verification_level: dict[str,int] = _find_verification_level(_data["verification_level"])
        self.default_msg_notifs: dict[str,int] = _find_notif_level(_data["default_message_notifications"])
        self.explicit_filter_lvl: dict[str,int] = _find_expl_level(_data["explicit_content_filter"])
        self.roles: list[Role] = [Role(role) for role in _data["roles"]]
        self.emojis: list[Emoji] = [Emoji(emoji) for emoji in _data["emoji"]]
        self.features: list[str] = _data["features"]
        self.mfa_level = "NONE" if _data["mfa_level"] == 0 else "ELEVATED"
        self.application_id: Optional[Snowflake] = Snowflake(_data["application_id"]) if _data.get("application_id") is not None else None
        self.system_channel_id: Optional[Snowflake] = Snowflake(_data["system_channel_id"]) if _data.get("system_channel_id") is not None else None
        self.system_channel_flags: list[dict[str,int]] = _get_system_flags(_data["system_channel_flags"])
        self.rules_channel_id: Optional[Snowflake] = Snowflake(_data["rules_channel_id"]) if _data.get("rules_channel_id") is not None else None
        self.max_presences: Optional[int] = _data["max_presences"] if _data.get("max_presences") is not None else None
        self.max_members: Optional[int] = _data["max_members"] if _data.get("max_members") is not None else None
        self.vanity_url_code: Optional[str] = _data.get("vanity_url_code")
        self.description: Optional[str] = _data.get("description")
        self.banner: Optional[str] = f"https://cdn.discordapp.com/banners/{self.id}/{_data["banner"]}.{"gif" if _data["banner"].startswith("a_") else "png"}" if _data.get("banner") else None  
        self.premium_tier: str = _get_nitro_rank(_data["premium_tier"])
        self.premium_subscription_count: int = _data["premium_subscription_count"]
        self.preferred_locale: str = Locales.__members__[_data["preferred_locale"].upper().replace("-","_")]
        self.public_upd_channel_id: Optional[Snowflake] = Snowflake(_data["public_updates_channel_id"]) if _data.get("public_updates_channel_id") is not None else None
        self.max_video_channel_users: Optional[int] = _data.get("max_video_channel_users")
        self.max_stage_video_channel_users: Optional[int] = _data.get("max_stage_video_channel_users")
        self.approximate_member_count: Optional[int] = _data.get("approximate_member_count")
        self.approximate_presence_count: Optional[int] = _data.get("approximate_presence_count")
        self.welcome_screen: Optional[WelcomeScreen] = WelcomeScreen(_data["welcome_screen"]) if _data["welcome_screen"] is not None else None
        self.stickers: Optional[list[Sticker]] = [Sticker(sticker) for sticker in _data["stickers"] if _data.get("stickers")]
        self.progress_bar_enabled: bool = _data["premium_progress_bar_enabled"]
        self.safety_alerts_channel_id = _data["safety_alerts_channel_id"]
        self.__channel_cache = []


    async def get_own_preview(self,bot: Bot):
        """
        **Co-routine** \n
        Retrieves the current guild's community preview.
        ## Parameters
        - bot - `discmoji.Bot`
          - Needed for permissions when retrieving data. (This is a minor inconvenience as we develop through later versions)
        ## Returns
        - **discmoji.GuildPreview**
           - The guild's preview.
        
        ## Raises
        - **DiscmojiRetrievalError**
           - Failed to retrieve community preview."""
        req = await bot.http.request("get",f"/guilds/{self.id}/preview")
        if req.status >= 400:
            raise DiscmojiRetrievalError(f"get_own_preview(bot: {bot})","Unspecified error occured while trying to retrieve current guild's preview.")    
        else:
            content = GuildPreviewMapper(_GuildPreviewPayload(req)).map()
            return content
    
    
    async def edit_guild(self,bot: Bot,**kwargs):
        """**Co-routine** \n
        Allows you to edit the guild's settings. 
        ## Parameters
         - bot - `discmoji.Bot`
           - Needed for authentication.
         - kwargs - `Any`
           - The fields in the guild you want to change. Check documentation for possible fields. (must be keyword arguments)
        ## Returns
        - **Self/discmoji.Guild**
          - The guild with changed settings/fields
        ## Raises
        - **Forbidden**
          - The bot isn't allowed to change a field.
        - **UnknownHTTPError**
          - An unspecified error occured."""
        req = await bot.http.request("patch",f"/guilds/{self.id}",kwargs=kwargs)
        #                                                          ^^^^^^^^^^^
        #                                                    Kinda weird, but will work
        if req.status == 403:
            raise Forbidden("Bot does not have permissions to edit field on current guild.")
        elif req.status >= 400:
            raise UnknownHTTPError(req.status,"Was not able to change a field on current guild.")
        else:
            for key,value in req.data.items():
                setattr(self,key,value)
            return self
            # returns self for good chaining
    
    
    async def delete_guild(self,bot: Bot):
        """*Dangerous Method* \n
        **Co-routine** \n
        Deletes the current guild. Beware, as this is **irreversible**. User must be owner.
        ## Parameters
        - bot - `discmoji.Bot`
          - Needed for authentication.
        ## Returns
        **None**
        ## Raises
        - **Forbidden**
          - User is not owner of the guild.
        - **UnknownHTTPError**
          - An unspecified error occured.
        """
        req = await bot.http.request("delete",f"/guilds/{self.id}")
        if req.status == 204:
            return
        if req.status == 403:
            raise Forbidden("Bot is not owner of the guild.")
        if req.status >= 400:
            raise UnknownHTTPError(req.status, "Was not able to delete the guild.")

    async def channels(self,bot: Bot):
      """
      **Co-routine** \n
      Returns the channels of the current guild.
      ## Parameters
      - bot - `discmoji.Bot`
         - Needed for authentication.
      
      ## Returns
      - `list[discmoji.Channel]`
        - A list of all the channels in the guild.
      ## Raises
      - **DiscmojiRetrievalError**
        - An unspecified error occured."""
      req = await bot.http.request("get",f"/guilds/{self.id}/channels")
      if req.status == 200:
        self.__channel_cache = [Channel(channel) for channel in req.data]
        return [Channel(channel) for channel in req.data]
        # raw data is sent into Channel since it's convenient and more readable
        # and won't cause any problems in the future
      else:
        raise DiscmojiRetrievalError(f"channels(bot: {bot})","Could not retrieve guild channels.")
    
    async def create_channel(self,bot: Bot,**kwargs):
      """
      **Co-routine** \n
      Creates a channel in the current guild.
      
      ## Parameters
      - bot - `discmoji.Bot`
        - Needed for authentication.
      - kwargs - `Any`
        - Fields you want to specify for the new channel.
      
      ## Returns
      **discmoji.Channel**
        - The new channel
      
      ## Raises
      **UnknownHTTPError**
        - An unspecified error occured.      
      """
      formatted_kwargs = []
      for value in kwargs.values():
        if value is not int | str | Snowflake | bool | dict:
          formatted_kwargs.append(value.__dict__)
        else:
          formatted_kwargs.append(value)
      req = await bot.http.request("post",f"/guilds/{self.id}/channels",kwargs=kwargs)
      if req.status <= 400:
        return ChannelMapper(_ChannelPayload(req.data)).map()
      else:
        raise UnknownHTTPError(req.status,"Failed to create channel.")
    
    async def modify_channel_position(self,bot: Bot,**kwargs):
      """Co-routine \n
      Modifies the provided channel's position on the guild's channel listing.
      
      ## Parameters
      - bot - `discmoji.Bot`
        - Needed for authentication.
      - kwargs - `Any`
        - Any keyword arguments that align with the field's name you want to change.
      
      ## Returns
      None
      
      ## Raises
      - UnknownHttpError
        - An unspecified error occurred."""
      req = await bot.http.request("patch",f"/guilds/{self.id}/channels",kwargs=kwargs)
      if req.status >= 400:
        raise UnknownHTTPError(req.status,"Failed to update channel position.")
    
    async def active_guild_threads(self,bot: Bot,with_thread_members: bool = False):
      """Co-routine \n
      Lists the active guild thread channels, with the members that chatted in them, if enabled.
      
      ## Parameters
      - bot - `discmoji.Bot`
        - Needed for authentication.
      - with_thread_members - `bool = False`
        - Specifies whether you want to get the thread members in that channel, or not. Defaults to false.
      
      ## Returns
      - `list[discmoji.Channel]`
        - The active guild threads.
      - `list[list[discmoji.Channel | discmoji.ThreadMember]]`
        - A list, containing 2 lists that contain the active guild threads, and the thread members respectively if enabled."""
      req = await bot.__http.request("get",f"/guilds/{self.id}/threads/active")
      if not with_thread_members:
        return [Channel(channel) for channel in req.data["threads"]]
      else:
        pass
      # will implement discmoji.ThreadMember soon
type Server = Guild 


"""Exceptions"""

class InternalDiscmojiException(Exception):
    """Something failed internally."""
    
    def __str__(self):
        return f"[DiscmojiError] Failed due to internal error. Cause: {self.args}"


class DiscmojiRetrievalError(Exception):
    """Something failed when trying to retrieve another thing."""
    def __init__(self,method: str,string: str):
        self.method = method
        self.message = string
    
    def __str__(self):
        return f"[DiscmojiError] Failed while running {self.method}: {self.message}"


class Forbidden(Exception):
    """The bot does not have the permissions to perform this action."""
    
    def __str__(self):
        return f"[403] Bot is forbidden to access resource. Check permissions for details."

class UnknownHTTPError(Exception):
    """An unknown http error has occured"""
    def __str__(self):
        return f"[Error Code: {self.args[0]}]: {self.args[1]}"


from ..c_payload import _ChannelPayload
from ..channel import Channel

class ChannelMapper:
    def __init__(self, data: _ChannelPayload):
        self.data = data
        
    
    def map(self):
        return Channel(self.data.__to_be_conv.data)



"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..snowflake import Snowflake
from .emoji import Emoji 
from .sticker import Sticker

class GuildPreview:
    """Represents a guild's preview, if bot is not in the guild, requires guild to be discoverable.
    
    ## Attributes
    - id - `discmoji.Snowflake`
      - The guild preview's id.
    - name - `str`
      - The guild's name.
    - icon - `str | None`
      - A formatted CDN link leading to the guild's icon.
    - splash - `str | None`
      - A formatted CDN link leading to the guild's splash.
    - discovery_splash - `str | None`
      - A formatted CDN link leading to the guild's discovery splash.
    - emojis - `list[discmoji.Emoji] | None`
      - A list of the guild's custom emojis.
    - features - `list[str]`
      - A list of all the guild's enabled features.
    - approx_member_count - `int`
      - An approximate of the current amount of members in the guild.
    - approx_presence_count - `int`
      - An approximate of the current amount of presences in the guild.
    - description - `str | None`
      - The guild's description.
    - stickers - `list[discmoji.Sticker]`
      - A list of the guild's custom stickers.
    """
    def __init__(self,dict: dict):
        self.id = Snowflake(dict["id"])
        self.name: str = dict["name"]
        self.icon = f"https://cdn.discordapp.com/icons/{self.id}/{dict["avatar"]}.{"gif" if dict["avatar"].startswith("a_") else "png"}" if dict.get("icon") else None
        self.splash = f"https://cdn.discordapp.com/splashes/{self.id}/{dict["splash"]}.{"gif" if dict["avatar"].startswith("a_") else "png"}" if dict.get("splash") else None
        self.discovery_splash =  f"https://cdn.discordapp.com/discovery-splashes/{self.id}/{dict.get("discovery_splash")}.{"gif" if dict.get("discovery_splash").startswith("a_") else "png"}" if dict.get("discovery_splash") else None
        self.emojis = [Emoji(emoji) for emoji in dict["emojis"]]
        self.features: list[str] = dict["features"]
        self.approx_member_count: int = dict["approximate_member_count"]
        self.approx_presence_count: int = dict["approximate_presence_count"]
        self.description = dict.get("description")
        self.stickers = [Sticker(sticker) for sticker in dict["stickers"]]


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..snowflake import Snowflake
from typing import Optional
from .overwrites import PermissionOverwrite
from datetime import datetime
from ..types import _get_channel_flags
from ..types import ForumTag as Tag
class Channel:
    """Represents a channel. Also the base class which all other Channel classes inherit from.
    ## Attributes
    - id - `discmoji.Snowflake`
      - ID of the channel.
    - guild_id - `Optional[discmoji.Snowflake]`
      - The ID of the guild this channel came from. May be none in some gateway payloads.
    - position - `Optional[int]`
      - The position of the channel in the guild's channel listing. May be none.
    - overwrites - `Optional[list[PermissionOverwrite]]`
      - A list of permission overwrites for this channel.
    - name - `Optional[str]`
      - The name of the channel.
    - topic - `Optional[str]`
      - Topic of this channel.
    - last_message_id - `Optional[discmoji.Snowflake]`
      - The last message in this channel's id, if exists.
    - rate_limit_per_user - `Optional[int]`
      - The slowmode (in seconds) for this channel, if applicable.
    - parent_id -  `Optional[discmoji.Snowflake]`
      - The category in which this channel is in if it's a guild and the text channel this thread was created if it's a thread.
    - last_pin_date - `Optional[datetime.datetime]`
      - A `datetime` object showing when the last message was pinned to this channel's pinboard, if any.
    - flags - `list`
      - A list of the names of the flags this channel has."""
    def __init__(self,_data: dict):
        self.id: Snowflake = Snowflake(_data["id"])
        self.guild_id: Optional[Snowflake] = Snowflake(_data["guild_id"]) if _data.get("guild_id") is not None else None
        self.position: Optional[int] = _data.get("position")
        self.overwrites: Optional[list[PermissionOverwrite]] = [PermissionOverwrite(overwrite) for overwrite in _data["permission_overwrites"]] if _data.get("permission_overwrites") else None
        self.name: Optional[str] = _data.get("name")
        self.topic: Optional[str] = _data.get("topic")
        self.last_message_id: Optional[Snowflake] = Snowflake(_data["last_message_id"]) if _data.get("last_message_id") is not None else None
        self.rate_limit_per_user: Optional[int] = _data.get("rate_limit_per_user")
        self.parent_id: Optional[Snowflake] = Snowflake(_data["parent_id"]) if _data.get("parent_id") is not None else None
        self.last_pin_date: Optional[datetime] = datetime.fromisoformat(_data["last_pin_timestamp"]) if _data.get("last_pin_timestamp") else None
        self.flags = _get_channel_flags(_data["flags"])
    

class VoiceChannel(Channel):
  """Represents a discord voice channel. Inherits from `discmoji.Channel`.
  ## Attributes
  - Note: This only records attributes that are specifically only in this class, which means all channel attributes will not appear here
  - bitrate - `int`
    - The bitrate of the channel.
  - user_limit - `int`
    - the limit to the amount of users in the same channel.
  - rtc_region - `discmoji.Snowflake`
    - The id of the voice region this channel is in.
  - quality_mode - Literal["Auto","Full"]
    - "Auto" if discord decides the video quality, else "Full", which is 720p.
    """
  def __init__(self,_data: dict):
    self.bitrate: int = _data["bitrate"]
    self.user_limit: int = _data["user_limit"]
    self.rtc_region: Snowflake = Snowflake(_data["rtc_region"])
    self.quality_mode = "Auto" if _data["video_quality_mode"] == 0 else "Full"



class ForumChannel(Channel):
  """Represents a discord forum/media channel. Inherits from `discmoji.Channel`.
  ## Attributes
  - Note: This only records attributes new to this class, not attributes from the base class.
  - available_tags - `list[discmoji.ForumTag]`
    - A list of thread tags that can be used for a thread. 
  - default_reaction_emoji - `discmoji.Snowflake`
    - The id of the default emoji to show as a reaction for every thread.
  - thread_rate_limit_per_user - `int`
    - the `rate_limit_per_user` for newly created threads.
  - default_sort_order - `Literal["Latest Activity","Creation Date"] | None`
    - The default way to sort threads. None if not configured."""
  def __init__(self,_data: dict):
    self.available_tags = [Tag(tag) for tag in _data["available_tags"]]
    self.default_reaction_emoji: Snowflake = Snowflake(_data["default_reaction_emoji"]["emoji_id"])
    self.thread_rate_limit_per_user: int = _data["default_thread_rate_limit_per_user"]
    if _data.get("default_sort_order") is not None:
      self.default_sort_order = "Latest Activity" if _data["default_sort_order"] == 0 else "Creation Date"
    else:
      self.default_sort_order = None
    self.__forum_layout_dict = {
      0 : "Not Set",
      1 : "List View",
      2 : "Gallery View"
    }
    self.layout = self.__forum_layout_dict[_data["default_forum_layout"]]


from ..gp_payload import _GuildPreviewPayload
from ..preview import GuildPreview    
    
class GuildPreviewMapper:    
    def __init__(self, dto: _GuildPreviewPayload):
        self.data = dto
    
    
    def map(self):
        return GuildPreview(self.data.__to_be_conv.data)


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import aiohttp
import websockets
import asyncio
import json
from enum import Enum, IntEnum
from typing import Any, Literal,Optional,Self
from .snowflake import Snowflake
import enum
from .exceptions import InternalDiscmojiException
import logging

class RequestBody:
    def __init__(self,response: aiohttp.ClientResponse):
        self.headers = response.headers
        self.data: dict = json.loads(asyncio.run(response.content.read()).decode())
        self.status = response.status


class WebsocketPayload:
    def __init__(self,response: Optional[websockets.Data],opcode: int ,data: dict | int | None):
        self.__serialized: dict = json.loads(response)
        self.opcode = self.__serialized.get("op") if self.__serialized.get("op") else opcode
        self.data = self.__serialized.get("d") if self.__serialized.get("data") else data
        self.seq = self.__serialized.get("s")
        self.event = self.__serialized.get("t")
    
    def jsonize(self) -> str:
        if isinstance(self.data,dict) and self.opcode not in range(0,31):
            self.data['op'] = 0 # type: ignore
            return json.dumps(self.data)
        elif isinstance(self.data, dict) and self.opcode in range(0,31):
            self.data['op'] = self.opcode # type: ignore
            # type ignore because on my vscode it keeps complaining about an undefined variable ;(
            return json.dumps(self.data)
    
class Locales(Enum):
    ID = "Indonesian"
    DA = "Danish"
    DE = "German"
    EN_GB = "UK English"
    EN_US = "US English"
    ES_ES = "Spanish"
    ES_419 = "Spanish, LATAM"
    FR = "French"
    HR = "Croatian"
    IT = "Italian"
    LT = "Lithuanian"
    HU = "Hungarian"
    NL = "Dutch"
    NO = "Norwegian"
    PL = "Polish"
    PT_BR = "Portuguese, Brazilian"
    RO = "Romanian"
    FI = "Finnish"
    SV_SE = "Swedish"
    VI = "Vietnamese"
    TR = "Turkish"
    CS = "Czech"
    EL = "Greek"
    BG = "Bulgarian"
    RU = "Russian"
    UK = "Ukrainian"
    HI = "Hindi"
    TH = "Thai"
    ZH_CN = "Chinese"
    JA = "Japanese"	
    ZH_TW = "Chinese, Taiwan"
    KO = "Korean"

class UserFlags(IntEnum):
    STAFF = 1 << 0
    PARTNER = 1 << 1
    HYPESQUAD = 1 << 2 
    BUG_HUNTER_1 = 1 << 3
    HOUSE_BRAVERY = 1 << 6
    HOUSE_BRILLIANCE = 1 << 7
    HOUSE_BALANCE = 1 << 8
    PREMIUM_EARLY_SUPPORTER = 1 << 9
    TEAM_PSEUDO_USER = 1 << 10
    BUG_HUNTER_2 = 1 << 14
    VERIFIED_BOT = 1 << 16
    VERIFIED_DEV = 1 << 17
    CERTIFIED_MOD = 1 << 18
    BOT_ONLY_HTTP = 1 << 19
    ACTIVE_DEV = 1 << 22

def _flags_parse(input: int) -> list[dict[str,int | Any]] | None:
    parsed = []
    bytesinput = bytes(input)
    for key,value in UserFlags.__members__.items():
        for num in bytesinput:
            if bytes(num) == value:
              parsed.append({key:value})
    if input == None:
        return
    return parsed

def _return_nitro_type(data: int) -> Literal["Nitro Classic","Nitro","Nitro Basic"] | None:
    if data == 0:
        return None
    if data == 1:
        return "Nitro Classic"
    if data == 2:
        return "Nitro"
    if data == 3:
        return "Nitro Basic"

class VerificationLevels(IntEnum):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    VERY_HIGH = 4
    
def _find_verification_level(input: int):
    for key,value in VerificationLevels.__members__.items():
        if value == input:
            return {key:value}

class DefaultMessageNotifLevel(Enum):
    ALL_MESSAGES = 0
    ONLY_MENTIONS = 1

def _find_notif_level(input: int):
    for key, value in DefaultMessageNotifLevel.__members__.items():
        if value == input:
            return {key:value}

class ExplicitContentFilter(Enum):
    DISABLED = 0
    MEMBERS_WITHOUT_ROLES = 1
    ALL_MEMBERS = 2

def _find_expl_level(input: int):
    for key, value in ExplicitContentFilter.__members__.items():
        if value == input:
            return {key:value}

class RoleTags:
    """Role metadata object.
    ## Attributes
    - bot_id - `Optional[discmoji.Snowflake]`
       - Snowflake representing the id of the bot associated with this role, can be None.
    - integration_id - `Optional[discmoji.Snowflake]`
       - Snowflake representing the id of the integration associated with this role, can be None.
    - premium_role - `Optional[bool]`
       - Indicates whether this is a booster role or not. (role given when boosting)
    - subscription_sku_listing - `Optional[discmoji.Snowflake]`
       - The id of this role's subscription SKU and listing, may be None.
    - available_for_purchase - `Optional[bool]`
       - Whether this role is a purchaseable role e.g whether this role is a role awarded through guild membership.
    - guild_connections - `Optional[bool]`
       - Whether this role is a guild's linked role."""
    def __init__(self,_data: dict):
        self.bot_id: Optional[Snowflake] = Snowflake(_data["bot_id"]) if _data.get("bot_id") is not None else None
        self.integration_id: Optional[Snowflake] = Snowflake(_data["integration_id"]) if _data.get("integration_id") is not None else None
        self.premium_role: Optional[bool] = _data.get("premium_subscriber") 
        self.subscription_sku_listing: Optional[Snowflake] = Snowflake(_data["subscription_listing_id"]) if _data.get("bot_id") is not None else None
        self.available_for_purchase: Optional[bool] = _data.get("available_for_purchase")
        self.guild_linked_role: Optional[bool] = _data.get("guild_linked_role")

class PermissionsBits(enum.IntFlag):
    """Enum used by permissions related functions to create or check permissions.
    Use this permissions enum with bitwise operators, like such:
    ```
    permissions = PermissionsBits.create_invites | PermissionsBits.kick_members | ...
    ```
    These are different from intents, do not mix them up. \n
    `|` - adds a permission \n
    `&` - checks if object that is applicable has that permission \n
    `~` -  checks if object that is applicable does NOT have that permission \n
    use these operators as such:
    ```python
    (perms & Permissions.create_invites) == 1 # true
    perms = PermissionsBits.Example | Permissions.OtherExample # 2304 (example, not accurate)
    ~PermissionsBits.LALALA == 21030 # true (example, not accurate)
    ```
    It is advised to use permission check/creator functions to do this work for you, since this may not work as expected.
    """
    create_invites = 1 << 0
    kick_members = 1 << 1
    ban_members = 1 << 2
    admin = 1 << 3
    manage_channels = 1 << 4
    manage_guild = 1 << 5
    reactions = 1 << 6
    audit_log = 1 << 7
    priority_speaker = 1 << 8
    stream = 1 << 9
    view_channel = 1 << 10
    send_messages = 1 << 11
    tts_messages = 1 << 12 
    manage_messages = 1 << 13
    embed_links = 1 << 14
    attach_files = 1 << 15
    read_message_history = 1 << 16 
    mention_everyone = 1 << 17
    external_emojis = 1 << 18
    guild_insights = 1 << 19
    connect = 1 << 20
    speak = 1 << 21
    mute_members = 1 << 22
    deafen_members = 1 << 23
    move_members = 1 << 24
    use_vad = 1 << 25
    change_nickname = 1 << 26
    manage_nicknames = 1 << 27
    manage_roles = 1 << 28
    manage_webhooks = 1 << 29
    guild_expressions = 1 << 30
    use_app_cmds = 1 << 31
    req_to_speak = 1 << 32
    events = 1 << 33
    threads = 1 << 34
    public_threads = 1 << 35
    private_threads = 1 << 36
    use_external_stickers = 1 << 37
    send_messages_threads = 1 << 38
    embedded_activities = 1 << 39
    moderate_members = 1 << 40
    creator_monetization_analytics = 1 << 41
    soundboard = 1 << 42
    external_sounds = 1 << 45
    voice_messages = 1 << 46
    polls = 1 << 49
    external_apps = 1 << 50
    

class Permissions:
    """Contains the result of functions that return permissions. 
    These classes can be used to create new permissions, through 'flipping' each one of the attributes. (e.g False to True)
    This provides a more abstracted interface to creating and checking permissions, than `discmoji.PermissionsBits`."""
    def __init__(self):
        self.create_invites = False
        """Creating invites"""
        self.kick_members = False
        """kicking members"""
        self.ban_members = False
        """banning members"""
        self.admin = False
        """administrator perms"""
        self.manage_channels = False
        """managing channels e.g deleting them, moving them"""
        self.manage_guilds = False
        """managing guilds e.g changing guild banner etc."""
        self.reactions = False
        """reaction related permissions"""
        self.audit_log = False
        """Viewing audit log events"""
        self.priority_speaker = False
        """Priority speaker in voice chats"""
        self.stream = False
        """Stream in voice chats"""
        self.view_channel = False
        """Viewing specific channel"""
        self.send_messages = False
        """Pretty self explanatory"""
        self.tts_messages = False
        """TTS messages, TTS standing for **T**ext **T**o **S**peech"""
        self.manage_messages = False
        """Allows you to delete messages"""
        self.embed_links = False
        """Allows links to automatically be embedded"""
        self.attach_files = False
        """Pretty self-explanatory here."""
        self.message_history = False
        """Message history"""
        self.everyone = False
        """Permission that allows you to mention everyone"""
        self.external_emojis = False
        """Allows you to use emojis from other servers that have been joined, if Nitro user."""
        self.insights = False
        """Allows you to view guild insights"""
        self.connect = False
        """Connecting in a VC (**V**oice **C**hannel)"""
        self.speak = False
        """Speaking in a VC"""
        self.mute_members = False
        """Muting members"""
        self.deafen_members = False
        """Deafening members in a VC"""
        self.move_members = False
        """Moving members to other channels"""
        self.use_vad = False
        """Allowing user in VC to use VAD. (**V**oice **A**ctivity **D**etection)"""
        self.change_nickname = False
        """Allows user to change own nickname"""
        self.manage_nicknames = False
        """Allows user to manage other user's nicknames"""
        self.manage_roles = False
        """Allows user to manage guild's roles."""
        self.manage_webhooks = False
        """Allows user to configure webhooks"""
        self.guild_expressions = False
        """Allows user to edit and delete guild stickers or emojis"""
        self.use_app_cmds = False
        """Allows user to use external application commands"""
        self.req_to_speak = False
        """Allows user to request to speak in a Stage channel"""
        self.events = False
        """Allows for deleting and scheduling events for discord servers."""
        self.threads = False
        """Allows for managing of threads"""
        self.public_threads = False
        """Allows user to create public threads"""
        self.private_threads = False
        """Allows user to create private threads"""
        self.external_stickers = False
        """Allows user to use stickers from other servers."""
        self.messages_in_threads = False
        """Allows user to send messages in threads"""
        self.embedded_activities = False
        """Allows for using activities"""
        self.moderate_members = False
        """Allows user to time out other users"""
        self.creator_monetization_analytics = False
        """Allows user to view role subscription insights."""
        self.soundboard = False
        """Usage of soundboard."""
        self.external_sounds = False
        """Usage of external sounds from other guilds."""
        self.voice_messages = False
        """Allowing of user to send voice messages."""
        self.polls = False
        """Allowing of user to create polls"""
        self.external_apps = False
        """Allows user to use external apps."""
        
    @classmethod
    def _convert_perms(cls,input: int) -> Self:
            """Internal method used to configure base permissions to liking."""
            return_class = cls()
            bytes_input = bytes(input)
            for key,item in PermissionsBits.__members__.items():
                if bytes_input & bytes(item) == True:
                    try:    
                        setattr(return_class,key,True)
                    except AttributeError as e:
                        raise InternalDiscmojiException(e.args)
            return return_class
    
    
    def _deconv_perms(self) -> int:
        """An internal utility function that turns the Permissions object back into an integer. """
        integer = 0
        for item,value in self.__dict__.items():
            if type(value) == function:
                continue
            else:
                if value:
                    integer += PermissionsBits.__members__[item]
        return integer
class SystemChannelFlags(Enum):
    SUPPRESS_JOIN_NOTIFS = 1 << 0
    SUPPRESS_PREMIUM_SUBSCRIPTIONS = 1 << 1
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS = 1 << 2
    SUPPRESS_JOIN_NOTIFICATION_REPLIES = 1 << 3
    SUPPRESS_ROLE_SUBSCRIPTION_PURCHASE_NOTIFICATIONS = 1 << 4
    SUPPRESS_ROLE_SUBSCRIPTION_PURCHASE_NOTIFICATION_REPLIES = 1 << 5

def _get_system_flags(input: int):
    byte_input = bytes(input)
    value_list = []
    for key,value in SystemChannelFlags.__members__.items():
        if byte_input & bytes(value):
            value_list.append({key:value})
    return value_list


def _get_nitro_rank(input: int):
    if input == 0:
        return None
    if input == 1:
        return "Tier 1"
    if input == 2:
        return "Tier 2"
    if input == 3:
        return "Tier 3"

logger = logging.getLogger("discmoji")
logger.addHandler(logging.NullHandler())
formatter = logging.Formatter(f"%(name) %(levelname) at : %(message)",datefmt="%I:%M")

def find_sticker_format_type(inp: int):
    if inp == 1:
        return "PNG"
    if inp == 2:
        return "APNG"
    if inp == 3:
        return "LOTTIE"
    if inp == 4:
        return "GIF"


class Opcodes(Enum):
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE_UPDATE = 3
    VOICE_STATE_UPDATE = 4
    RESUME = 6
    RECONNECT = 7
    REQUEST_GUILD_MEMBERS = 8
    INVALID_SESSION = 9
    HELLO = 10
    ACK = 11
    REQ_SOUNDBOARD_SOUNDS = 31


class IntentsBits(enum.IntFlag):
    """A less abstracted interface than `discmoji.BotIntents` for making intents.
    This enum is meant to be used with bitwise operators. \n
    -| - adds an intent \n
    & - checks if bot has that intent \n
    ~ - checks if bot does not have that intent \n
    ```python
    # examples
    perms = IntentsBits.Example | IntentsBits.OtherExample
    (perms & IntentsBits.OtherExample) == True # true
    ~perms == 0 # true
    ```
    Note that this returns an integer, not a BotIntents class.
    """
    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1
    GUILD_MODERATION = 1 << 2
    GUILD_EXPRESSIONS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    DM = 1 << 12
    DM_REACTIONS = 1 << 13
    DM_TYPING = 1 << 14
    MESSAGE_CONTENT = 1 << 15
    GUILD_SCHEDULED_EVENTS = 1 << 16
    AUTOMOD_CONFIG = 1 << 20
    AUTOMOD_EXECUTION = 1 << 21
    GUILD_MESSAGE_POLLS = 1 << 24
    DM_POLLS = 1 << 25

class ChannelFlags(enum.IntFlag):
    PINNED = 1 << 1
    REQUIRE_TAG = 1 << 4
    HIDE_MEDIA_DOWNLOAD_OPTIONS = 1 << 15
    

def _get_channel_flags(inp: int):
    byteinp = bytes(inp)
    returns = []
    for item,value in ChannelFlags.__members__.items():
        if byteinp & bytes(value):
            returns.append(item)
    return returns


class ForumTag:
    """Represents a tag that is able to be applied to a thread in a forum or media channel.
    ## Attributes
    - id - `discmoji.Snowflake`
      - The id of the tag.
    - name - `str`
      - the name of the tag.
    - moderated - `bool`
      - Whether this tag has to be added by a user with the `MANAGE_THREAD` permission or not.
    - emoji_id - `Optional[Snowflake]`
      - The id of the emoji that's on the tag (if exists)
    - emoji_name - `Optional[str]`
      - The name of the emoji that's on the tag (if exists)"""
    def __init__(self,_data: dict) -> None:
        self.id: Snowflake = Snowflake(_data["id"])
        self.name: str = _data["name"]
        self.moderated: bool = _data["moderated"]
        self.emoji_id: Optional[Snowflake] = Snowflake(_data["emoji_id"]) if _data.get("emoji_id") is not None else None
        self.emoji_name: Optional[str] = _data["emoji_name"] if _data.get("emoji_name") else None


from ..gm_payload import _GuildMemberPayload
from ..guild_member import GuildMember

class GuildMemberMapper:
    def __init__(self, dto: _GuildMemberPayload):
        self.data = dto
    
    
    def map(self):
        return GuildMember(self.data.__to_be_conv.data)


from ..rt_payload import _RoleTagsPayload
from ...types import RoleTags


class RoleTagMapper:
    def __init__(self, dto: _RoleTagsPayload):
        self.data = dto

    
    def map(self):
        return RoleTags(self.data.__to_be_conv.data)


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from .snowflake import Snowflake
from .types import Locales, UserFlags, _flags_parse, _return_nitro_type
from .decoration import AvatarDecoration

class User:
    """Class that represents a discord user.
      ## Attributes
        - id - `discmoji.Snowflake`
          - The id of the User.
        - username - `str`
          - The username of the User.
        - discriminator - `str`
          - The discriminator of the User.
        - global_name - `str | None`
          - The global display name of the user. May not exist.
        - avatar - `str | None`
          - Contains a formatted CDN link that can be used to retrieve the avatar. May not exist.
        - bot - `bool | None`
          - An attribute indicating if the user that this model is representing is an application. May not exist.
        - system - `bool | None`
          - An attribute which indicates whether the user this model is representing is a discord system user/ is part of the urgent message system. (see documentation on User object)
        - mfa - `bool | None`
          - Indicates if the User has multi-factor authentication enabled or not.
        - banner - `str | None`
          - contains a formatted CDN link that can be used to retrieve the banner. May not exist.
        - accent_color - `str | None`
          - contains a hex representation of a color (the "0x" prefix is removed). May be None.
        - locale - `str | None`
          - Indicates the User's chosen language option. May not exist.
        - flags - `list[dict[str,int]] | None`
          - Contains the user's (formatted) flags. May not exist.
        - nitro_type - `Literal["Nitro","Nitro Basic","Nitro Classic"] | None`
          - Contains any one of these literals, depending on the type. May not exist.
        - public_flags - `list[dict[str,int]] | None`
          - Contains the user's (formatted) public flags. May not exist.
        - avatar_decoration - `discmoji.AvatarDecoration`
          - Contains the User's chosen decoration's CDN link.
          """
    def __init__(self,data: dict[str,dict | str | int]):
        self.id: Snowflake = Snowflake(data["id"])
        self.username: str = data["username"]
        self.discriminator: str = data["discriminator"]
        self.global_name: str | None = data.get("global_name")
        self.avatar: str | None = f"https://cdn.discordapp.com/avatar/{self.id}/{data.get("avatar")}.{"gif" if data.get("avatar").startswith("a_") else "png"}" if data.get("avatar") is not None else None
        # avatar is just formatted str for now, as the asset object is made
        self.bot: bool | None = data.get("bot")
        self.system: bool | None = data.get("system")
        self.mfa: bool | None = data.get("mfa_enabled")
        self.banner: str | None = f"https://cdn.discordapp.com/banners/{self.id}/{data.get("banner")}.{"gif" if data.get("banner").startswith("a_") else "png"}" if data.get("banner") is not None else None
        self.accent_color: str | None = hex(data.get("accent_color")).removeprefix("0x") if data.get("accent_color") is not None else None
        self.locale: str | None = Locales.__dict__.get(data.get("locale").upper().replace("-","_")) if data.get("locale") is not None else None
        self.flags = _flags_parse(UserFlags,data.get("flags"))
        self.nitro_type = _return_nitro_type(data.get("premium_type"))
        self.public_flags = _flags_parse(UserFlags,data.get("flags"))
        self.avatar_decoration = AvatarDecoration(data.get("avatar_decoration_data"))
        
        
        


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..snowflake import Snowflake
from typing import Optional

class WelcomeScreenChannel:
    """Represents a Welcome Screen's channels.
    ## Attributes
    - channel_id - `discmoji.Snowflake`
      - The channel id.
    - description - `str`
      - The description provided for this in the welcome screen.
    - emoji_id - `Optional[discmoji.Snowflake]`
      - The emoji id for this welcome screen channel's emoji, if it is custom.
    - emoji_name - `Optional[str]`
      - The name for the emoji if custom, the unicode emoji if standard, and null if no emoji is set."""
    def __init__(self,_data: dict):
        self.channel_id = Snowflake(_data["channel_id"])
        self.description: str = _data["description"]
        self.emoji_id: Optional[Snowflake] = _data.get("emoji_id")
        self.emoji_name: Optional[str] = _data.get("emoji_name")

class WelcomeScreen:
    """Represents a Guild's welcome screen.
    ## Attributes
    - description - `Optional[str]`
     - description for the server, shown in the welcome screen.
    - welcome_channels - `list[discmoji.WelcomeScreenChannel]`
     - List of `WelcomeScreenChannel`'s that are a part of this server's welcome screen."""
    def __init__(self,dict: dict):
        self.description: Optional[str] = dict.get("description")
        self.welcome_channels = [WelcomeScreenChannel(channel) for channel in dict["welcome_channels"]]
        


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this event notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from .types import IntentsBits
from typing import Self
from .exceptions import InternalDiscmojiException
class BotIntents:
  """Represents your bot's gateway intents. 
  You can turn intents on and off using the corresponding attribute, and flipping its value to True or False. 
  Everything is set false by default."""
  def __init__(self):
        self.guilds = False
        """The following events are enabled with this intent: 
  - `GUILD_CREATE`
  - `GUILD_UPDATE`
  - `GUILD_DELETE`
  - `GUILD_ROLE_CREATE`
  - `GUILD_ROLE_UPDATE`
  - `GUILD_ROLE_DELETE`
  - `CHANNEL_CREATE`
  - `CHANNEL_UPDATE`
  - `CHANNEL_DELETE`
  - `CHANNEL_PINS_UPDATE`
  - `THREAD_CREATE`
  - `THREAD_UPDATE`
  - `THREAD_DELETE`
  - `THREAD_LIST_SYNC`
  - `THREAD_MEMBER_UPDATE`
  - `THREAD_MEMBERS_UPDATE`
  - `STAGE_INSTANCE_CREATE`
  - `STAGE_INSTANCE_UPDATE`
  - `STAGE_INSTANCE_DELETE`"""
        self.guild_members = False
        """(Privileged Intent) The following events are enabled with this privileged intent:
  - `GUILD_MEMBER_ADD`
  - `GUILD_MEMBER_UPDATE`
  - `GUILD_MEMBER_REMOVE`
  - `THREAD_MEMBERS_UPDATE`"""
        self.guild_moderation = False
        """The following events are enabled with this intent:
  - `GUILD_AUDIT_LOG_ENTRY_CREATE`
  - `GUILD_BAN_ADD`
  - `GUILD_BAN_REMOVE`"""
        self.guild_expressions = False
        """The following events are enabled with this intent.
  - `GUILD_EMOJIS_UPDATE`
  - `GUILD_STICKERS_UPDATE`
  - `GUILD_SOUNDBOARD_SOUND_CREATE`
  - `GUILD_SOUNDBOARD_SOUND_UPDATE`
  - `GUILD_SOUNDBOARD_SOUND_DELETE`
  - `GUILD_SOUNDBOARD_SOUNDS_UPDATE`"""
        self.guild_integrations = False
        """The following events are enabled with this intent.
  - `GUILD_INTEGRATIONS_UPDATE`
  - `INTEGRATION_CREATE`
  - `INTEGRATION_UPDATE`
  - `INTEGRATION_DELETE`"""
        self.guild_webhooks = False
        """The following event is enabled with this intent:
        - `WEBHOOKS_UPDATE`"""
        self.guild_invites = False
        """ The following events are enabled with this intent: 
  - `INVITE_CREATE`
  - `INVITE_DELETE`"""
        self.guild_voice_states = False
        """The following events are enabled with this intent:
  - `VOICE_CHANNEL_EFFECT_SEND`
  - `VOICE_STATE_UPDATE`"""
        self.guild_presences = False
        """(Privileged) The following events are enabled with this privileged intent:
        - `PRESENCE_UPDATE`"""
        self.guild_messages = False
        """The following events are enabled with this intent:
  - `MESSAGE_CREATE`
  - `MESSAGE_UPDATE`
  - `MESSAGE_DELETE`
  - `MESSAGE_DELETE_BULK`"""
        self.guild_message_reactions = False
        """The following events are enabled with this intent:  
  - `MESSAGE_REACTION_ADD`
  - `MESSAGE_REACTION_REMOVE`
  - `MESSAGE_REACTION_REMOVE_ALL`
  - `MESSAGE_REACTION_REMOVE_EMOJI` """
        self.guild_message_typing = False
        """The following event is enabled with this intent:
        - `TYPING_START`"""
        self.dm = False
        """The following events are enabled with this intent:
  - `MESSAGE_CREATE`
  - `MESSAGE_UPDATE`
  - `MESSAGE_DELETE`
  - `CHANNEL_PINS_UPDATE`"""
        self.dm_reactions = False
        """The following events are enabled with this intent:
  - `MESSAGE_REACTION_ADD`
  - `MESSAGE_REACTION_REMOVE`
  - `MESSAGE_REACTION_REMOVE_ALL`
  - `MESSAGE_REACTION_REMOVE_EMOJI`"""
        self.dm_typing = False
        """The following events are enabled with this intent:
          - `TYPING_START`"""
        self.message_content = False
        """(Privileged) No events are enabled with this intent, but instead, makes you eligible to recieve message content. Necessary for prefix commands to work."""
        self.guild_scheduled_events = False
        """The following events are enabled with this intent:
  - `GUILD_SCHEDULED_EVENT_CREATE`
  - `GUILD_SCHEDULED_EVENT_UPDATE`
  - `GUILD_SCHEDULED_EVENT_DELETE`
  - `GUILD_SCHEDULED_EVENT_USER_ADD`
  - `GUILD_SCHEDULED_EVENT_USER_REMOVE`"""
        self.automod_config = False
        """The following events are enabled with this intent:
  - `AUTO_MODERATION_RULE_CREATE`
  - `AUTO_MODERATION_RULE_UPDATE`
  - `AUTO_MODERATION_RULE_DELETE` """
        self.automod_execution = False
        """The following event is enabled with this intent:
          - `AUTO_MODERATION_ACTION_EXECUTION`"""
        self.guild_message_polls = False
        """The following events are enabled with this intent:
  - `MESSAGE_POLL_VOTE_ADD`
  - `MESSAGE_POLL_VOTE_REMOVE`"""
        self.dm_polls = False
        """The following events are enabled with this intent:
  - `MESSAGE_POLL_VOTE_ADD`
  - `MESSAGE_POLL_VOTE_REMOVE`"""
  
      
  @classmethod
  def _conv_intents(cls,input: int) -> Self:
        # literally just a copy and paste from Permissions functions
        """Converts from a raw integer representation of intents to self."""
        return_class = cls()
        bytes_input = bytes(input)
        for key,item in IntentsBits.__members__.items():
                if bytes_input & bytes(item):
                    try:    
                        setattr(return_class,key,True)
                    except AttributeError as e:
                        raise InternalDiscmojiException(e.args)
        return return_class
  
  def _deconv_intents(self):
        """An internal utility function that turns the Intents object back into an integer. """
        integer = 0
        for item,value in self.__dict__.items():
            if type(value) == function:
                continue
            else:
                if value:
                    integer += IntentsBits.__members__[item.upper()]
        return integer


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ._http import HttpManager
import asyncio
from .intents import BotIntents
import websockets
from .types import WebsocketPayload,Opcodes,logger
from contextlib import asynccontextmanager
from typing import Self
from random import uniform
import json

class DiscordWebsocket:
    def __init__(self,http: HttpManager,intents: BotIntents):
        self.ws = None
        self.token = http.token
        self.intents = intents._deconv_intents()
        self.interval: int = None
        self.seq: int | None = None 
        self.url: str = None
    
    @asynccontextmanager
    # huge credit to graingert on discord!
    async def initiate_connection(cls,http: HttpManager,intents: BotIntents):
        url = await http.request('get','/gateway/bot',True).data["url"]
        async with websockets.connect(url) as ws:
                cls.ws = ws
                try:
                    goingtobeyield: Self = cls(ws,http,intents)
                    yield goingtobeyield
                except websockets.ConnectionClosedError:
                    raise RuntimeError("Gateway closed connection.")
                    
    
    async def _recieve_latest(self):
        result: dict = json.loads(await self.ws.recv())
        self.seq = result.get("s")
        # .get because sequence field may be none at times
        return WebsocketPayload(opcode=result["op"],data=result["d"])

    async def first_heart_beat(self):
        msg = WebsocketPayload(opcode=Opcodes.HEARTBEAT,data=None)
        await self.ws.send(msg.jsonize())
    
    async def resume(self):
        msg = WebsocketPayload(opcode=Opcodes.RESUME,data={
            "token": self.token,
            "session_id": self.url,
            "seq": self.seq
        })
        await self.ws.send(msg.jsonize())
        recv: dict = json.loads(await self.ws.recv())
        if recv.get("op") != 9:
            logger.info(f"Successfully resumed connection with gateway. Payload: \n {recv}")
        else:
            logger.fatal(f"Invalid Session. Report this to devs with payload! {recv}")
            
    
    async def _heart_beat_loop(self,seq: int):
        while True:
            recv = json.loads(await self.ws.recv())
            logger.info(f"Payload recieved: \n {recv}")
            await asyncio.sleep(self.interval)
            # placed above since no wait during sending reconnect event
            if recv["op"] == Opcodes.RECONNECT:
                await self.resume()
            if recv["op"] == Opcodes.HEARTBEAT:
                send = WebsocketPayload(opcode=recv["op"],data=recv["d"])
                await self.ws.send(send.jsonize())
            msg = WebsocketPayload(opcode=Opcodes.HEARTBEAT,data=seq)
            await self.ws.send(msg.jsonize())
    
    async def _handshake(self):
        recv = await self._recieve_latest()
        if recv.opcode == Opcodes.HELLO:
            self.interval = recv.data["heartbeat_interval"]
            await asyncio.sleep(uniform(0,1))
            await self.first_heart_beat()
            asyncio.create_task(self._heart_beat_loop(self.seq))
        if recv.opcode == Opcodes.ACK:
            logger.info(f"Initiated Heartbeat with gateway at interval of {self.interval / 1000} s.")
        identify = WebsocketPayload(opcode=2,data = {
            "token": self.token,
            "properties": {
                "os": "linux",
                "browser": "discmoji",
                "device": "discmoji"
            },
            "intents": self.intents
        })
        await self.ws.send(identify.jsonize())
        recv = await self._recieve_latest()
        
        if recv.opcode == Opcodes.DISPATCH:
            logger.info(f"Successfully established connection to gateway at session id: {recv.data["session_id"]}")
            logger.info(f"Bot Name: {recv.data["application"]["name"]}")
            self.url = recv.data["resume_gateway_url"]
        
        
            


from ..g_payload import _GuildPayload
from ..guild import Guild

class GuildMapper:
    def __init__(self, dto: _GuildPayload):
        self.data = dto
    
    
    def map(self):
        return Guild(self.data.__to_be_conv.data)


"""MIT License

Copyright (c) 2024 mojidev-py

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from typing import Optional
from ..user import User
from .guild import Guild
from ..snowflake import Snowflake
import datetime
from ..decoration import AvatarDecoration
class GuildMember:
    """Represents a member in a guild.
    ## Attributes
    - user - `discmoji.User | None`
        - contains the user object of the guild member. Will not be present in prefix command context.
    - nick - `Optional[str]`
        - contains the nickname of the guild member.
    - avatar - `Optional[str]`
        - contains the formatted CDN link of the guild member's avatar.
    - banner - `Optional[str]`
        - contains the formatted CDN link of the guild member's banner.
    - roles - `Optional[str]`
        - contains the role id's of the guild member. Role objects can be retrieved through `discmoji.Guild.get_role()`.
    - joined_at - `datetime.datetime`
        - contains the join date of the guild member, converted from ISO8601 format.
    - premium_since - `Optional[datetime.datetime]`
        - contains the time elapsed since the guild member first boosted the guild, converted from ISO8601 format.
    - deafened - `bool`
        - Indicates whether the guild member is deafened in a voice channel or not.
    - muted - `bool`
        - Indicates whether the guild member is muted in a voice channel or not.
    - pending - `Optional[bool]`
        - Indicates whether the guild member has passed membership screening or not.
    - timeout_until - `Optional[datetime.datetime]`
        - Contains the date that the guild member's timeout ends, if applicable, else is None.
    - avatar_decoration - `Optional[discmoji.AvatarDecoration]`
        - Contains the Avatar Decoration object, if it exists."""
    def __init__(self,_dict: dict[str,str | int | dict | None],__binded_guild: Guild):
        self.user: User | None = User(_dict.get("user")) if _dict.get("user") is not None else None 
        self.nick: Optional[str] = _dict.get("nick")
        self.avatar: Optional[str] = f"https://cdn.discordapp.com/guilds/{__binded_guild.id}/users/{self.user.id if isinstance(self.user,User) else None}/avatars/{_dict.get("avatar")}.{"gif" if _dict.get("avatar").startswith("a_") else "png"}"
        self.banner: Optional[str] = f"https://cdn.discordapp.com/guilds/{__binded_guild.id}/users/{self.user.id if isinstance(self.user,User) else None}/banners/{_dict.get("avatar")}.{"gif" if _dict.get("avatar").startswith("a_") else "png"}"
        self.roles: list[Snowflake] = [Snowflake(roleid) for roleid in _dict["roles"]]
        self.joined_at: datetime.datetime = datetime.datetime.fromisoformat(_dict["joined_at"])
        self.premium_since: Optional[datetime.datetime] = datetime.datetime.fromisoformat(_dict.get("premium_since"))
        self.deafened: bool = _dict.get("deaf")
        self.muted: bool = _dict.get("mute")
        self.pending: Optional[bool] = _dict.get("pending")
        self.timeout_until: Optional[datetime.datetime] = datetime.datetime.fromisoformat(_dict.get("communication_disabled_until")) if datetime.datetime.fromisoformat(_dict.get("communication_disabled_until")) > datetime.datetime.now() or _dict.get("communication_disabled_until") is not None else None
        self.avatar_decoration: Optional[AvatarDecoration] = AvatarDecoration(_dict.get("avatar_decoration_data")) if _dict.get("avatar_decoration_data") is not None else None
        


from ..r_payload import _RolePayload
from ..roles import Role


class RoleMapper:
    def __init__(self,dto: _RolePayload):
        self.data = dto
    
    def map(self):
        return Role(self.data.__to_be_conv.data)